import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from num2words import num2words
import re
import mplcursors
import math

def check_kinetic_data(kinetic_data,bold_and_red=True,print_report=True):
    """
    Checks whether the kinetic data loaded has the right format for auto_vtna and creates a report \
    string with relevant information and warnings. The function checks that the kinetic_data \
    variable is a dictionary with dataframe elements, that the column titles are consistent across \
    experiment sheets, that there are more than 2 columns per experiment dataframe and that all \
    concentration and time values are numerical. To load the kinetic data correctly, first \
    create an excel file with the concentration profiles from each experiment in separate sheets.\
    Each sheet should start with a time column with a suitable column title in cell A1, followed by concentration \
    values over time for different reaction species, each with column titles referring to the compound name \
    in the first row of sheet (B1, C1, D1 etc.): kinetic_data=pd.read_excel('file_name.xlsx',sheet_name=None). \
    Alternatively, the data from each experiment can be loaded as separate csv files using \
    pd.read_csv('filename.csv') and then collected as a dictionary manually. This takes more time.\
    """
    # Define functions used to check if for numerical values in lists.
    def all_elements_not_convertible_to_float(my_list):
        return all(not isinstance(element, (int, float)) and not (isinstance(element, str) 
        and element.replace('.', '', 1).isdigit()) for element in my_list)
    def check_numerical(values):
        return [isinstance(val, (int, float)) for val in values]
    def count_convertible_to_float(my_list):
        return sum(isinstance(element, (int, float)) or (isinstance(element, str) and 
        element.replace('.', '', 1).isdigit()) for element in my_list)
    # Define function to remove escape sequences from the result string. 
    def remove_escape_sequences(input_string):
        # Define the pattern for escape sequences
        escape_sequence_pattern = re.compile(r'\x1b\[[0-9;]*m')
        # Use sub() to replace all matches with an empty string
        cleaned_string = escape_sequence_pattern.sub('', input_string)
        return cleaned_string
    error=None
    tick_symbol='\u2713'
    # Begin defining the report string and its first section. 
    report='Kinetic data check report.\n--------------------------\n\033[1mCheck 1: Correct \
loading of kinetic data from Excel.\033[0m\n'
    # Ensure that kinetic_data is a dictionary, end the check if it isn't. 
    has_empty_dataframes = any(df.empty for df in kinetic_data.values())
    if type(kinetic_data)!=dict or has_empty_dataframes:
        report=report+"\033[91mOBS:\033[0m The kinetic data is not loaded correctly. It should \
be a dictionary containing the different experiment datasets as dataframes\nMake sure that there \
are no empty sheets and that the excel file contains one experiment dataset per\nsheet with column \
titles for time and concentration of reaction species and load the data like this: \
pd.read_excel('file_name.xlsx',sheet_name=None)\n"
        report=report+f'\033[1mConclusion: the kinetic_data variable is not ready for use with the\
auto_vtna package.\033[0m'
        # Remove escape sequences from the result string if bol_and_red set to False. 
        if bold_and_red==False:
            report=remove_escape_sequences(report)
        if print_report:
            print(report)
        return report
    # Show that it has been confirmed that kinetic_data is a dictionary. 
    report=report+f'  {tick_symbol} The kinetic data is loaded as a dictionary.\n'
    # Check that each key:value pair in kinetic_data contains a pandas dataframe object.
    type_elements=[]
    for i in kinetic_data.keys():
        type_elements.append(str(type(kinetic_data[i])))
    if all(i=="<class 'pandas.core.frame.DataFrame'>" for i in type_elements):
        report=report+f'  {tick_symbol} Each element associated a key in the kinetic data dictionary \
is a pandas dataframe object.\n'
    # If each key:value pair in kinetic_data isn't a pandas dataframe object, end the check.
    else: 
        report=report+"\033[91mOBS:\033[0m Some of the elements in the kinetic data dictionary are \
not pandas dataframe objects. This will need to be fixed.\nMake sure the excel file contains one \
experiment dataset per\n sheet with column titles for time and concentration of reaction species \
and load the data like this: pd.read_excel('file_name.xlsx',sheet_name=None)\n"
        report=report+f'\033[1mConclusion: the kinetic_data variable is not ready for use with \
the auto_vtna package.\033[0m'
        # Remove escape sequences from the result string if bol_and_red set to False. 
        if bold_and_red==False:
            report=remove_escape_sequences(report)
        if print_report:
            print(report)
        return report
    # List the experiment labels from each dataframe in the kinetic_data dictionary.
    no_experiments=len(list(kinetic_data.keys()))
    # Create a list containing lists of column titles from each dataframe. 
    column_titles=[]
    for exp in kinetic_data.keys():
        column_titles.append(kinetic_data[exp].columns.tolist())
    # Create a list containing the number of columns for each dataframe. 
    no_columns=[]
    columns=[]
    for exp in kinetic_data.keys():
        no_columns.append(len(kinetic_data[exp].columns))
        columns.append(list(kinetic_data[exp].columns))
    # Check if there is only one dataframe in kinetic_data (not VTNA compatible).
    if len(no_columns)==1:
        error=True
        report=report+'\033[91mOBS:\033[0m The kinetic data only contains data from one \
experiment. This is not enough for VTNA.\n'
    # Check if the number of columns in each dataframe is the same. 
    if all(i==no_columns[0] for i in no_columns):
        report=report+f'  {tick_symbol} The dataframe element(s) in the kinetic data \
dictionary all contain {no_columns[0]} columns.\n'
        # Check if every dataframe contains only one column (not VTNA compatible).
        if 1 in no_columns:
            report=report+'\033[91mOBS:\033[0m The dataframe elements in the kinetic \
data dictionary contains only one column.\
            This is insufficient for VTNA.\n'
            report=report+f'\033[1mConclusion: the kinetic data is not ready for use \
with the auto_vtna package.\033[0m'
            # Remove escape sequences from the result string if bol_and_red set to False. 
            if bold_and_red==False:
                report=remove_escape_sequences(report)
            if print_report:
                print(report)
            return report
    # Check if the column titles are the same across every kinetic experiment.
    report=report+'\033[1mCheck 2: Column title consistency.\033[0m\n'
    # Check if the columns from each dataframe are the same length. 
    if all(i==no_columns[0] for i in no_columns):
        column_titles = np.array(column_titles)
        # Check if all the reaction species concentration columns are the same.
        are_equal = np.all(column_titles == column_titles[0, :], axis=0)
        no_columns=len(column_titles[0])
        if all(are_equal):
            report=report+f'  {tick_symbol} All experiment datasets have the same column titles.\n'
        else:
            # Update the error flag.
            error=True
            # Flag each column title that is not consistent across dataframes. 
            for i in range(no_columns):
                if are_equal[i]==False:
                    report=report+f'\033[91mOBS:\033[0m The {num2words(i+1, to="ordinal_num")} \
column title is different across experiments: {column_titles[:,i]} \n'
    # If the number of columns is inconsistent across dataframes, update the error flag.
    else:
        error=True
        # Add an OBS statement to the report string to informa that the number of columns is 
        # inconsistent and list the column titles of each dataframe. 
        report=report+ f'\033[91mOBS:\033[0m'
        report=report+f' the number of columns in the dataframe objects in \
kinetic_data are not consistent.\nThe column titles are:\n'
        for i in range(len(column_titles)):
            report=report+f"  - {list(kinetic_data.keys())[i]}: {column_titles[i]}\n"
    # Check if all kinetic data dataframes contain only numerical values.
    counter=0
    report=report+'\033[1mCheck 3: Numerical datapoints.\033[0m \n'
    for exp in kinetic_data.keys():
        for column in list(kinetic_data[exp].columns):
            # Create a list which is True for all the values in the column that are numerical values.
            num_column=[isinstance(val, (int, float)) and not math.isnan(val) for val in list(kinetic_data[exp][column])]
            # If no values in the column are non-numerical and therefore all values in num_column True,
            # Go to the next column. 
            if all(num_column):
                continue
            else:
                # If there are non-numerical values in the column, update the error flag and 
                # add an OBS statement to the report string. 
                error=True
                counter=counter+1
                report=report+ '\033[91mOBS:\033[0m'
                report=report+f' Row(s) number {[index+2 for index,value in enumerate(num_column) if value==False]}\
 in column "{column}" in dataset "{exp}" contains values which are not numerical. Needs to be fixed.\n'.replace('[','').replace(']','')
    # If the counter is still 0, this means all dataframes are fully numerical. 
    if counter==0:
        report=report+f'  {tick_symbol} Every experiment dataset contains only numerical\
values.\n'
    counter=0
    report=report+'\033[1mCheck 4: Positive datapoints.\033[0m \n'
    for exp in kinetic_data.keys():
        for column in list(kinetic_data[exp].columns):
            # Create a list which is True for all the values in the column that are not negative numerical values.
            any_negative_values = list(kinetic_data[exp][column].apply(lambda s: pd.to_numeric(s, errors='coerce') < 0))
            # If no values in the column are negative and therefore all values in any_negative_values True,
            # Go to the next column. 
            if not any(any_negative_values):
                continue
            else:
                # If there are non-numerical values in the dataframe, update the error flag and 
                # add an OBS statement to the report string. 
                error=True
                counter=counter+1
                report=report+ '\033[91mOBS:\033[0m'
                report=report+f' Row(s) number {[index+2 for index,value in enumerate(any_negative_values) if value==True]}\
 in column "{column}" in dataset "{exp}" contains negative values. Needs to be fixed.\n'.replace('[','').replace(']','')
    # If the counter is still 0, this means all dataframes are fully numerical. 
    if counter==0:
        report=report+f'  {tick_symbol} No datasets contain negative time or concentration values.\n'
    # Check if the column titles are the same across every kinetic experiment.
    
    # Proceed to check whether the time values in each sheet are monotonic. 
    report=report+'\033[1mCheck 5: Time column monotonicity.\033[0m\n'
    counter=0
    for exp in kinetic_data.keys():
        time_array=kinetic_data[exp][kinetic_data[exp].columns[0]].to_numpy()
        if not np.all(time_array[1:] > time_array[:-1]):
            report=report+ f'\033[91mOBS:\033[0m The time values in dataset {exp} are not strictly monotonically increasing. Needs to be fixed.\n'
            error=True
            counter=counter+1
    if counter==0:
        report=report+f'  {tick_symbol} The time column (first column) of the dataset in every sheet is monotonically increasing.\n'
    # Check if the column titles are the same across every kinetic experiment.
    report=report+f'\033[1mCheck 6: Time and concentration columns.\033[0m \n'
    check_for_floats=0
    # Create a flattened column title list. 
    columns_flat=[]
    for i in columns:
        columns_flat=columns_flat+i
    # Check if all column titles are strings. 
    if all(type(i)==str for i in columns_flat):
        report=report+f'  {tick_symbol} All dataframe elements in the kinetic data dictionary \
has column titles which are strings.\n' 
    else: 
        for exp_no in range(len(list(kinetic_data.keys()))):
            # Ensure that all column titles aren't strings that can be converted to numerical values.
            if all_elements_not_convertible_to_float(columns[exp_no]):
                continue
            else:
                # If some column titles are numerical, add an OBS statement to the report string.
                check_for_floats=1
                report=report+f"\033[91mOBS:\033[0m {count_convertible_to_float(columns[exp_no])}\
of the column titles of dataframe {list(kinetic_data.keys())[exp_no]} are numerical values.\n"
                # Update error as numerical column titles reflect incorrect kinetic_data format.
                error=True
    # If no column title is a numerical value, add this to the report string. 
    if check_for_floats==0:
        report=report+f'  {tick_symbol} The column titles of every experiment dataset are all \
strings.\n'
    # Define the time column titles from each experiment dataframe. 
    time_column_titles=[i[0] for i in columns]
    # Check if the time column title is consistent across dataframes. 
    if all(i==time_column_titles[0] for i in time_column_titles):
        time_column_titles=time_column_titles[0]
    # Make the user aware of the title of the columns that will be interpreted as time values. 
    report=report+f"\033[91mNB:\033[0m Make sure that the time columns are the first columns in \
each dataframe in kinetic_data. Currently, the first column titles are: {time_column_titles}.\n"
    # If no errors have been flagged, generate a positive conclusion (kinetic_data is ready)
    if error==None:
        report=report+f'\033[1mConclusion:\033[0m\033[0m The kinetic data has been loaded \
correctly if the time axis label is correct.'
    # Generate a conclusion warning if one or more errors in kinetic_data errors have been detected. 
    else:
        report=report+f"\033[91m\033[1mConclusion:\033[0m The kinetic data has not been loaded \
correctly.Make sure that the kinetic data Excel spreadsheet is organised correctly and try again \
using kinetic_data=pd.read_excel('file_name.xlsx',sheet_name=None)."
    # Remove escape sequences from the result string if bol_and_red set to False. 
    if bold_and_red==False:
        report=remove_escape_sequences(report)
    if print_report:
        print(report)
    return report

def make_VTNA_selection(data,normalised_species='NS',output_species='OS'):
    """
    Creates a default VTNA selection dictionary structure with keys for each \
    experiment name as well as the normalised_species and output_species. The \
    dictionary can then be copied by the user and information about the desired \
    VTNA filled in such as:
    - RM: Sub dictionary containing each experiment dataset (RM stands for reaction \
    mixture) with an associated list of omitted reaction species by list index. 
    - normalised_species: Sub dictionary containing information about which reaction \
    species' concentration profiles will be used to normalise the time axis. Each \
    normalised species has an associated reaction order value for time normalisation. 
    - output_species: The reaction species whose concentration profile will define \
    y-axis. 
    The keyword arguments normalised_species and output_species can be defined to \
    generate a VTNA selection dictionary with the relevant keys.
    """
    selection_dict={}
    selection_dict['RM']={}
    for i in data.keys():
        selection_dict['RM'][i]={}
        selection_dict['RM'][i]['omissions']=None
        if i==normalised_species:
            selection_dict['normalised_species']={}
            selection_dict['normalised_species'][i]=1
        if i==output_species:
            selection_dict['output_species']=output_species
    if output_species=='OS':
        selection_dict['output_species']={}
        selection_dict['output_species']='OS'
    if normalised_species=='NS':
        selection_dict['normalised_species']={}
        selection_dict['normalised_species']['NS']=1
    return selection_dict

def initial_concs(data):
    """
    Creates a Pandas DataFrame showing the initial concentrations of every reaction species across each \
    experiment. Requires that each experiment DataFrame in "data" have the same column titles.
    """
    # Obtain a list of the experiment names
    experiments=list(data.keys())
    # Obtain a list containing the reaction species with concentration profiles in the kinetic dataset.
    reaction_species=data[experiments[0]].columns[1:]
    # Loop through each experiment and obtain the intial concentration of each reaction species.
    initial_conc_dict={'Reaction species':reaction_species}
    for j in experiments:
        initial_conc_dict[f'{j} init. conc.']=data[j].iloc[0,1:].to_numpy()
    return pd.DataFrame(initial_conc_dict)

def plot_data(data, ylim=None, y_unit='M', t_unit=None, fig_size_scaler=1,plot_mode='together',
    legend_outside=False, DP_scaler=1,linewidth=1):
    """
    Inputs the kinetic data "data" that has been imported from an excel file using pd.read_excel() \
    as a dictionary of DataFrames containing the concentration profiles for each different excess \
    experiment. Plots the concentration profiles of each each experiment in separate plots unless \
    plot_mode is set to 'together' or 'scrollable'.
    Args:
        data (dict): Kinetic data as a dictionary of Pandas dataframes.
        ylim (float, optional): Y axis upper cutoff value so that Y∈[0, ylim]. 
        y_unit (str, optional): Concentration unit to be included in the Y axis label. 
        t_unit (str, optional): Time unit to be included in the X axis label. 
        fig_size_scaler (float or int, optional): Scaler to adjust the size of the graphs by the \
        selected factor.
        plot_mode (str): Either set to "scrollable" to generate a single graph which updates \
        to the next or previous graph by clicking right or left arrow keys, or 'together' to \
        plot graphs for each experiment together. If plot_mode is set to something else, \
        graphs will pop-up one after the other. 
        legend_outside (bol): Set to False by default to show legend inside the axis of the plot. \
        Can be set to True to show the legend outside the axis (top right). 
        DP_scaler (float, int): Factor by which the datapoints of the plots is scaled from a standard \
        value of 6. 
        linewidth (float, int): The linewidth parameter used to generate plots. Is by default set to 1. 
    """
    # Make the plot mode 'separate' if there is only one experiment in the dataset. 
    if len(list(data.keys()))==1:
        plot_mode='separate'
    # Set the datapoint size scaler to 1 if an incorrect value has been assigned.
    if DP_scaler==None or isinstance(DP_scaler, str):
        DP_scaler=1
    if plot_mode=='together':
        addition=0
        if legend_outside:
            addition=1
        # Check how many rows and columns there should be based on the number of dataframes in data.
        n,m = len(list(data.keys()))+addition,3
        if n>6:
            m=4
        num_cols = min(m, n)
        num_rows = (n + num_cols - 1) // num_cols
        # Add an exception if there are 9 plot objects:
        if len(list(data.keys()))+addition==9:
            num_cols,num_rows=3,3
        # Create a single figure and subplots with specified number of rows and columns.
        fig, axs = plt.subplots(nrows=num_rows, ncols=num_cols, figsize=(7 * 0.45*fig_size_scaler * num_cols, 5 * 0.6*num_rows * fig_size_scaler),constrained_layout=True)
        for i, (rm_name, rm_data) in enumerate(data.items()):
            # Check if there's only one row
            if num_rows == 1:  
                ax = axs[i % num_cols]  # Use modulo to handle indexing in a 1D array
            else:
                row, col = divmod(i, num_cols)
                ax = axs[row, col]
            time = rm_data[rm_data.columns[0]]
            ax.set_title(rm_name)
            # Plot kinetic profiles for each reaction species for a given reaction mixture data set
            for j in rm_data.columns[1:]:
                ax.plot(time, rm_data[j], label=j, linestyle='-', marker='o',markersize=6*DP_scaler,linewidth=linewidth)
            ax.set_xlabel(rm_data.columns[0])
            # Add a unit to the time axis title is specified. 
            if t_unit is not None:
                ax.set_xlabel(f'{rm_data.columns[0]} / {t_unit}')
            ax.set_ylabel(f'Concentration / {y_unit}')
            # Define y-axis range up to ylim if defined.
            if ylim is not None:
                ax.set_ylim(0, ylim)
            # Add a legend if legend_outise is not set to true:
            if not legend_outside:
                ax.legend()
        # Add an extra axis object if legend_outside is set to True. 
        if legend_outside:
            i=i+1
            if num_rows == 1:  # Check if there's only one row
                ax = axs[i % num_cols]  # Use modulo to handle indexing in a 1D array
                handles, labels = axs[0].get_legend_handles_labels()
            else:
                row, col = divmod(i, num_cols)
                ax = axs[row, col]
                handles, labels = axs[0,0].get_legend_handles_labels()
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            ax.spines['left'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.legend(handles, labels, loc='center')
        # Remove empty subplots if there are any
        for i in range(len(data)+addition, num_cols * num_rows):
            fig.delaxes(axs.flatten()[i])
        plt.show()
    elif plot_mode=='scrollable':
        # Define a class for creating scrollable plots
        class ScrollablePlot:
            def __init__(self, data_dict, ylim=None, y_unit='M', t_unit=None, 
                         fig_size_scaler=1,legend_outside=legend_outside,DP_scaler=1,linewidth=1):
                # Extract data and labels from the given dictionary
                data_1 = [data_dict[i] for i, j in data_dict.items()]
                data_labels = list(data_dict.keys())
                # Initialize instance variables
                self.data = data_1
                self.labels = data_labels
                self.current_plot_index = 0
                self.ylim = ylim
                self.y_unit = y_unit
                self.t_unit = t_unit
                self.fig_size_scaler = fig_size_scaler
                self.legend_outside=legend_outside
                self.DP_scaler=DP_scaler
                self.linewidth=linewidth
                # Create a Matplotlib figure and axis
                self.fig, self.ax = plt.subplots()
                # Create a note to inform user of scrolling feature.
                note_text = 'NB: Use left or right arrows to scroll to other experiments.'
                self.fig.text(0.5, 0.015, note_text, ha='center', color='gray')
                # Plot the initial data
                self.plot_current_data()
                # Enable hovering using mplcursors
                mplcursors.cursor(hover=True)
                # Connect the key press event to the corresponding method
                self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)

            def plot_current_data(self):
                # Clear the current axis
                self.ax.clear()
                # Get the data and label for the current plot
                current_data = self.data[self.current_plot_index]
                current_label = self.labels[self.current_plot_index]
                # Extract time and concentration data
                t = current_data.columns[0]
                time_column = current_data[t]
                # Plot each concentration profile
                for i in current_data.keys():
                    if i == t:
                        continue
                    self.ax.plot(time_column, current_data[i], label=i, marker='o',
                                 markersize=6*self.DP_scaler,linewidth=self.linewidth)
                # Set the figure size
                self.fig.set_size_inches(7 *0.8* self.fig_size_scaler, 5 *0.8* self.fig_size_scaler)
                # Set plot title
                self.ax.set_title(f'Concentration profiles in {current_label}')
                # Set x-axis label
                if self.t_unit is not None:
                    self.ax.set_xlabel(f'Time / {self.t_unit}')
                else:
                    self.ax.set_xlabel(t)
                # Set y-axis label
                self.ax.set_ylabel(f'Concentration / {self.y_unit}')
                # Set y-axis limit if specified
                if self.ylim is not None and isinstance(self.ylim, (float, int)):
                    self.ax.set_ylim(0, self.ylim)
                # Display legend
                if self.legend_outside:
                    # Calculate the width of the legend dynamically based on the longest text element
                    legend = self.ax.legend()
                    # Obtain the legend width in display units. 
                    legend_bbox = legend.get_window_extent()
                    legend_width = legend_bbox.width
                    # Obtain the figure width in display units.
                    fig_box = plt.gcf().get_window_extent()
                    fig_width_display = fig_box.width
                    # See by what fraction the figure width needs to be extended for outside legend.
                    legend_width_scaled=legend_width/fig_width_display
                    # Scale the right argument according to the legend width
                    self.fig.subplots_adjust(right=1 - legend_width_scaled)
                    # Place the legend outside the axis
                    self.ax.legend(loc='center left',bbox_to_anchor=(1, 0.8))
                    # Adjust figure size to increase width according to the outside legend width
                    fig_size = plt.gcf().get_size_inches()
                    fig_size[0] = fig_size[0] + fig_size[0]*(legend_width_scaled) # Increase width by the padding
                    plt.gcf().set_size_inches(fig_size)
                else:
                    self.ax.legend()
                # Adjust layout to prevent overlapping of titles and labels
                plt.tight_layout()
                # Redraw the plot
                plt.draw()
            def on_key_press(self, event):
                # Handle key presses for navigating between plots
                if event.key == 'right':
                    self.next_plot()
                elif event.key == 'left':
                    self.previous_plot()
            def next_plot(self):
                # Move to the next plot and update the display
                self.current_plot_index = (self.current_plot_index + 1) % len(self.data)
                self.plot_current_data()
            def previous_plot(self):
                # Move to the previous plot and update the display
                self.current_plot_index = (self.current_plot_index - 1) % len(self.data)
                self.plot_current_data()
        scrollable_plot = ScrollablePlot(data, ylim=ylim, y_unit=y_unit, t_unit=t_unit, 
        fig_size_scaler=fig_size_scaler,legend_outside=legend_outside,DP_scaler=DP_scaler,linewidth=linewidth)
        plt.show()
    else:
            # Identify the first key in the data dictionary
        RM1=list(data.keys())[0]
        # Identify the time column unit for the first reaction mixture data set
        t=data[RM1].columns[0]
        for i in data.keys():
            plt.figure(figsize=(7*0.8*fig_size_scaler, 5*0.8*fig_size_scaler))
            time=data[i][t]
            # plot kinetic profiles for each reaction species for a given reaction mixture data set
            for j in data[i].keys():
                if j==t:
                    continue
                plt.plot(time,data[i][j],label=j,linestyle='-',marker='o',markersize=6*DP_scaler,linewidth=linewidth)
                plt.title(i)
                plt.xlabel(t)
                if t_unit!=None:
                    plt.xlabel(f'{t} / {t_unit}')
                plt.ylabel(f'Concentration / {y_unit}')
            # Define y-axis range up to ylim if defined. 
            if legend_outside:
                # Calculate the width of the legend dynamically based on the longest text element
                legend = plt.legend()
                # Obtain the legend width in display units. 
                legend_bbox = legend.get_window_extent()
                legend_width = legend_bbox.width
                # Obtain the figure width in display units.
                fig_box = plt.gcf().get_window_extent()
                fig_width_display = fig_box.width
                # See by what fraction the figure width needs to be extended for outside legend.
                legend_width_scaled=legend_width/fig_width_display
                # Scale the right argument according to the legend width
                plt.subplots_adjust(right=1 - legend_width_scaled)
                # Place the legend outside the axis
                plt.legend(loc='center left', bbox_to_anchor=(1, 0.8))
                # Adjust figure size to increase width according to the outside legend width
                fig_size = plt.gcf().get_size_inches()
                fig_size[0] = fig_size[0] + fig_size[0]*(legend_width_scaled) # Increase width by the padding
                plt.gcf().set_size_inches(fig_size)
            else:
                plt.legend()
            if ylim!=None:
                plt.ylim(0,ylim)
            #Use plt.show to complete plot and move to the next reaction mixture data set
            plt.tight_layout()
            plt.show()

def plot_data_together(data,species,ylim=None,y_unit='mM',fig_size_scaler=1,t_unit=None,
    legend_outside=False,DP_scaler=1,linewidth=1):
    """
    Inputs the kinetic data "data" that has been imported from an excel file using Pd.read_excel() \
    as a dictionary of DataFrames containing the concentration profiles for each different excess \
    experiment. Creates a plot with the concentration profiles of the selected reaction species from \
    each experiment. 
    Args:
        data (dict): Kinetic data as a dictionary of Pandas dataframes.
        species (list): List containing the reaction species whose concentration profiles \
        will be plotted. 
        ylim (float or int, optional): Y axis upper cutoff value so that Y∈[0,ylim]. 
        y_unit (str, optional): Concentration unit to be included in the Y axis label.
        fig_size_scaler (float or int, optional): Scaler to adjust the size of the graphs by the \
        selected factor.
        t_unit (str): Unit of the time axis. Gives a "/ " followed by the string as an addition to \
        the column title of the time column in the inputted kinetic dataset. 
        legend_outside (bol): Set to False by default to show legend inside the axis of the plot. \
        Can be set to True to show the legend outside the axis (top right). 
        DP_scaler (float, int): Factor by which the datapoints of the plots is scaled from a standard \
        value of 6. 
        linewidth (float, int): The linewidth parameter used to generate plots. Is by default set to 1. 
    """
    # Identify the first key in the data dictionary
    RM1=list(data.keys())[0]
    # Identify the time column unit for the first reaction mixture data set
    t=data[RM1].columns[0]
    # Set the datapoint size scaler to 1 if an incorrect value has been assigned.
    if DP_scaler==None or isinstance(DP_scaler, str):
        DP_scaler=1
    t_label=t
    if t_unit!=None:
        t_label=t_label+f' / {t_unit}'
    fig, ax = plt.subplots(figsize=(7*0.9*fig_size_scaler, 5*0.9*fig_size_scaler))
    for i in data.keys():
        time=data[i][t]
        # plot kinetic profiles for the chosen species for each reaction mixture data set
        for j in data[i].keys():
            if j==t or j not in species:
                continue
            ax.plot(time,data[i][j],label=f'{i}, {species}',linestyle='-',marker='o',
                    markersize=6*DP_scaler,linewidth=linewidth)
            ax.set_title(f'concentration profiles for {species}')
            ax.set_xlabel(t_label)
            ax.set_ylabel(f'concentration / {y_unit}')
        # Define y-axis range up to ylim if defined.
        if ylim!=None:
            ax.set_ylim(0,ylim)
    if legend_outside:
        # Calculate the width of the legend dynamically based on the longest text element
        legend = plt.legend()
        # Obtain the legend width in display units. 
        legend_bbox = legend.get_window_extent()
        legend_width = legend_bbox.width
        # Obtain the figure width in display units.
        fig_box = plt.gcf().get_window_extent()
        fig_width_display = fig_box.width
        # See by what fraction the figure width needs to be extended for outside legend.
        legend_width_scaled=legend_width/fig_width_display
        # Scale the right argument according to the legend width
        plt.subplots_adjust(right=1 - legend_width_scaled)
        # Place the legend outside the axis
        plt.legend(loc='center left',bbox_to_anchor=(1, 0.8))
        # Adjust figure size to increase width according to the outside legend width
        fig_size = plt.gcf().get_size_inches()
        fig_size[0] = fig_size[0] + fig_size[0]*(legend_width_scaled) # Increase width by the padding
        plt.gcf().set_size_inches(fig_size)
    else:
        ax.legend()
    plt.tight_layout()
    plt.show()

def plot_data_MB(data, species, stochiometry, scaler=1, ylim=None, t_unit=None, y_unit='mM', fig_size_scaler=1,
                 plot_mode='together',legend_outside=False,DP_scaler=1,linewidth=1):
    """
    Inputs the kinetic data "data" that has been imported from an excel file using Pd.read_excel() \
    as a dictionary of DataFrames containing the concentration profiles for each different excess \
    experiment. Plots the concentration profiles of each each experiment with the mass balance relative \
    to time=0 of the chosen reaction species and their stochiometries. The graphs can be plotted in one \
    figure (plot_mode = 'together'), in separate pop-up figures (plot_mode='separate') or in a single \
    figure that can be updated upon 
    Args:
        data (dict): Kinetic data as a dictionary of Pandas dataframes.
        species (list): List containing the reaction species whose concentration profiles \
        will be used to calculate the mass balance. 
        stociometry (list): List containing the stociometries corresponding to the species \
        list.
        ylim (float or int, optional): Y axis upper cutoff value so that Y∈[0,ylim]. 
        y_unit (str, optional): Concentration unit to be included in the Y axis label.
        fig_size_scaler (float or int, optional): Scaler to adjust the size of the graphs by the \
        selected factor.
        plot_mode (str): Either set to "scrollable" to generate a single graph which updates \
        to the next or previous graph by clicking right or left arrow keys, or 'together' to \
        plot graphs for each experiment together. If plot_mode is set to something else, \
        graphs will pop-up one after the other. 
        legend_outside (bol): Set to False by default to show legend inside the axis of the plot. \
        Can be set to True to show the legend outside the axis (top right). 
        DP_scaler (float, int): Factor by which the datapoints of the plots is scaled from a standard \
        value of 6. 
        linewidth (float, int): The linewidth parameter used to generate plots. Is by default set to 1. 
    """
    # Make the plot mode 'separate' if there is only one experiment in the dataset. 
    if len(list(data.keys()))==1:
        plot_mode='separate'
    # Find the title of the time column of the experiment datasets
    RM1 = list(data.keys())[0]
    t = data[RM1].columns[0]
    # Set the datapoint size scaler to 1 if an incorrect value has been assigned.
    if DP_scaler==None or isinstance(DP_scaler, str):
        DP_scaler=1
    MB_title_string = ''
    # Define the title of the mass balance axis
    for mol in range(len(species)):
        if mol != 0 and species[mol] == species[-1]:
            MB_title_string = MB_title_string + 'and '
        MB_title_string = MB_title_string + f'{species[mol]}, '
    if len(species) == 1:
        MB_title_string = f'{species[0]}. '
    if plot_mode=='together':
        addition=0
        if legend_outside:
            addition=1
        # Check how many rows and columns there should be based on the number of dataframes in data.
        n,m = len(list(data.keys()))+addition,3
        if n>6:
            m=4
        num_cols = min(m, n)
        num_rows = (n + num_cols - 1) // num_cols
        # Add an exception if there are 9 plot objects:
        if len(list(data.keys()))+addition==9:
            num_cols,num_rows=3,3
        # Create a single figure and subplots with specified number of rows and columns.
        fig, axes = plt.subplots(nrows=num_rows, ncols=num_cols, figsize=(7.5 * 0.45*fig_size_scaler * num_cols, 4.5 * 0.6*num_rows * fig_size_scaler),constrained_layout=True)
        axes = axes.flatten() 
        for idx, (i, ax) in enumerate(zip(data.keys(), axes)):
            Concs = []
            largest_conc = []
            time = data[i][t]
            # Create a curve for each reaction species
            for j in data[i].keys():
                if j == t:
                    continue
                ax.plot(time, data[i][j], label=j, linestyle='-', marker='o', markersize=6*DP_scaler,linewidth=linewidth)
                ax.set_title(i)
                ax.set_xlabel(t)
                ax.set_ylabel(f'concentration / {y_unit}')
                largest_conc.append(max(data[i][j].to_numpy()))
                if t_unit is not None:
                    ax.set_xlabel(f'{t} / {t_unit}')
                if j == t or j not in species:
                    continue
                # Add concentration*stochiometry array for calculation of mass balance over time
                Concs.append(data[i][j].to_numpy() * stochiometry[species.index(j)])
            # Use Concs to create mass balance sum for each time point
            sums = np.zeros(len(Concs[0]))
            for i in Concs:
                sums = sums + i
            # Use sums to find the % mass balance at each time relative to the initial value.
            # If the first sum is 0, use the final sum instead. (could occur if only product(s) are chosen)
            if sums[0]>0:
                MB = sums / sums[0] * 100 * scaler
            else:
                MB = sums / sums[-1] * 100 * scaler
            ax2 = ax.twinx()
            ax2.set_ylabel(f'Mass balance for:\n {MB_title_string[:-2]}', color="purple")
            ax2.plot(time, MB, label="Mass balance %", linestyle='-', marker='o', color="purple",markersize=6*DP_scaler)
            ax2.plot([0, max(time)], [100, 100], linestyle='--', marker='', color='purple')
            ax2.legend(loc='upper center')
            # Ensure appropriate y limits with space for legends.
            ax2.set_ylim(0, max(MB) + 28)
            if ylim is not None and type(ylim) in [float,int]:
                ax.set_ylim(0, ylim)
            else:
                ax.set_ylim(0, max(largest_conc) + max(largest_conc) * 0.15)
            if not legend_outside:
                ax.legend()
        ax = axes[idx+1]
        handles, labels = axes[0].get_legend_handles_labels()
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])
        ax.legend(handles, labels, loc='center left')
        for i in range(len(data)+addition, num_cols * num_rows):
          fig.delaxes(axes.flatten()[i])
        # Adjust layout to prevent overlap
        plt.tight_layout()
        plt.show()
    elif plot_mode=='scrollable':
        class ScrollablePlot:
            def __init__(self, data_dict,species, stochiometry,y_unit='M',t_unit=None,fig_size_scaler=1,
                         ylim=None,scaler=1,legend_outside=False,DP_scaler=1,linewidth=1):
                data_1=[]
                for i,j in data_dict.items():
                    data_1.append(data_dict[i])
                data_labels=list(data_dict.keys())
                self.data = data_1
                self.labels = data_labels
                self.current_plot_index = 0
                self.y_unit=y_unit
                self.t_unit=t_unit
                self.ylim=ylim
                self.fig_size_scaler=fig_size_scaler
                self.species=species
                self.stochiometry=stochiometry
                self.MB_title_string=MB_title_string
                self.legend_outside=legend_outside
                self.DP_scaler=DP_scaler
                self.linewidth=linewidth
                mplcursors.cursor(hover=True)
                self.fig, self.ax = plt.subplots(figsize=(7 *0.8* fig_size_scaler, 5 *0.8* fig_size_scaler))
                # Create a note to inform user of scrolling feature.
                note_text = 'NB: Use left or right arrows to scroll to other experiments.'
                self.fig.text(0.5, 0.015, note_text, ha='center', color='gray')
                self.ax2=self.ax.twinx()
                self.plot_current_data()
                self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)

            def plot_current_data(self):
                self.ax.clear()
                self.ax2.clear()
                current_data = self.data[self.current_plot_index]
                current_label = self.labels[self.current_plot_index]
                t=current_data.columns[0]
                time_column=current_data[t]
                Concs=[]
                largest_conc=[]
                for j in current_data.keys():
                    if j==t:
                        continue
                    self.ax.plot(time_column,current_data[j],label=j,linestyle='-',marker='o',
                                 markersize=6*self.DP_scaler,linewidth=linewidth)
                    self.ax.set_title(f'Concentration profiles in {current_label}')
                    self.fig.set_size_inches(7*0.8*self.fig_size_scaler, 5*0.8*self.fig_size_scaler)
                    self.ax.set_xlabel(t)
                    self.ax.set_ylabel(f'concentration / {self.y_unit}')
                    largest_conc.append(max(current_data[j].to_numpy()))
                    if self.t_unit!=None:
                        self.ax.set_xlabel(f'Time / {self.t_unit}')
                    if j==t or j not in self.species:
                        continue
                    # Add concentration*stochiometry array for calculation of mass balance over time
                    Concs.append(current_data[j].to_numpy()*self.stochiometry[self.species.index(j)])
                # Use Concs to create mass balance sum for each time point
                sums=np.zeros(len(Concs[0]))
                for i in Concs:
                    sums=sums+i
                # Use sums to find the % mass balance at each time relative to the intial value.
                MB=sums/sums[0]*100* scaler
                self.ax2.set_ylabel(f'Mass balance for:\n {self.MB_title_string[:-2]}',color="purple")
                self.ax2.yaxis.set_label_position('right')
                self.ax2.plot(time_column,MB,label="Mass balance %",linestyle='-',marker='o',color="purple",markersize=6*self.DP_scaler)
                self.ax2.plot([0,max(time_column)],[100,100],linestyle='--',marker='',color='purple')
                self.ax2.legend(loc='upper center')
                # Ensure appropriate y limits with space for legends. 
                self.ax2.set_ylim(0,max(MB)+28)
                if self.ylim!=None and type(self.ylim) in [float,int]:
                    self.ax.set_ylim(0,self.ylim)
                else:
                    self.ax.set_ylim(0,max(largest_conc)+max(largest_conc)*0.15)
                if self.legend_outside:
                    # Calculate the width of the legend dynamically based on the longest text element
                    legend = self.ax.legend()
                    # Obtain the legend width in display units. 
                    legend_bbox = legend.get_window_extent()
                    legend_width = legend_bbox.width
                    # Obtain the figure width in display units.
                    fig_box = plt.gcf().get_window_extent()
                    fig_width_display = fig_box.width
                    # See by what fraction the figure width needs to be extended for outside legend.
                    legend_width_scaled=legend_width/fig_width_display
                    # Scale the right argument according to the legend width
                    self.fig.subplots_adjust(right=1 - legend_width_scaled-0.07)
                    # Place the legend outside the axis
                    legend_position_scaler=fig_size_scaler**0.3 if fig_size_scaler<1 else 1
                    self.ax.legend(loc='center left',bbox_to_anchor=(1.14/legend_position_scaler, 0.8))
                    # Adjust figure size to increase width according to the outside legend width
                    fig_size = plt.gcf().get_size_inches()
                    fig_size[0] = fig_size[0] + fig_size[0]*(legend_width_scaled) # Increase width by the padding
                    plt.gcf().set_size_inches(fig_size)
                else:
                    self.ax.legend()
                plt.tight_layout()
                plt.draw()

            def on_key_press(self, event):
                if event.key == 'right':
                    self.next_plot()
                elif event.key == 'left':
                    self.previous_plot()

            def next_plot(self):
                self.current_plot_index = (self.current_plot_index + 1) % len(self.data)
                self.plot_current_data()

            def previous_plot(self):
                self.current_plot_index = (self.current_plot_index - 1) % len(self.data)
                self.plot_current_data()
        scrollable_plot = ScrollablePlot(data,species,stochiometry,fig_size_scaler=fig_size_scaler,
        t_unit=t_unit,y_unit=y_unit,scaler=scaler,ylim=ylim,legend_outside=legend_outside,linewidth=linewidth, DP_scaler=DP_scaler)
        plt.show()
    else: 
        # Create a twin plot for each experiment dataset
        for i in data.keys():
            Concs=[]
            largest_conc=[]
            fig,ax = plt.subplots(figsize=(7*0.8*fig_size_scaler, 5*0.8*fig_size_scaler))
            fig_size = plt.gcf().get_size_inches()
            time=data[i][t]
            # Create a curve for each reaction species
            for j in data[i].keys():
                if j==t:
                    continue
                ax.plot(time,data[i][j],label=j,linestyle='-',marker='o',
                        markersize=6*DP_scaler,linewidth=linewidth)
                ax.set_title(i)
                ax.set_xlabel(t)
                ax.set_ylabel(f'concentration / {y_unit}')
                largest_conc.append(max(data[i][j].to_numpy()))
                if t_unit!=None:
                    plt.xlabel(f'{t} / {t_unit}')
                if j==t or j not in species:
                    continue
                # Add concentration*stochiometry array for calculation of mass balance over time
                Concs.append(data[i][j].to_numpy()*stochiometry[species.index(j)])
            if legend_outside:
                # Calculate the width of the legend dynamically based on the longest text element
                legend = ax.legend()
                # Obtain the legend width in display units. 
                legend_bbox = legend.get_window_extent()
                legend_width = legend_bbox.width
                # Obtain the figure width in display units.
                fig_box = plt.gcf().get_window_extent()
                fig_width_display = fig_box.width
                # See by what fraction the figure width needs to be extended for outside legend.
                legend_width_scaled=legend_width/fig_width_display
                # Scale the right argument according to the legend width
                fig.subplots_adjust(right=1 - legend_width_scaled-0.07)
                # Place the legend outside the axis
                legend_position_scaler=fig_size_scaler**0.3 if fig_size_scaler<1 else 1
                ax.legend(loc='center left',bbox_to_anchor=(1.14/legend_position_scaler, 0.8))
                # Adjust figure size to increase width according to the outside legend width
                fig_size = plt.gcf().get_size_inches()
                fig_size[0] = fig_size[0] + fig_size[0]*(legend_width_scaled) # Increase width by the padding
                plt.gcf().set_size_inches(fig_size)
            else:
                ax.legend()
            # Use Concs to create mass balance sum for each time point
            sums=np.zeros(len(Concs[0]))
            for i in Concs:
                sums=sums+i
            # Use sums to find the % mass balance at each time relative to the intial value.
            MB=sums/sums[0]*100*scaler
            ax2=ax.twinx()
            ax2.set_ylabel(f'Mass balance for:\n {MB_title_string[:-2]}',color="purple")
            ax2.plot(time,MB,label="Mass balance %",linestyle='-',marker='o',color="purple",markersize=6*DP_scaler)
            ax2.plot([0,max(time)],[100,100],linestyle='--',marker='',color='purple')
            ax2.legend(loc='upper center')
            # Ensure appropriate y limits with space for legends. 
            ax2.set_ylim(0,max(MB)+28)
            if ylim!=None and type(ylim) in [float,int]:
                ax.set_ylim(0,ylim)
            else:
                ax.set_ylim(0,max(largest_conc)+max(largest_conc)*0.15)
            fig.subplots_adjust(right=0.85)
            plt.show()




import numpy as np
import time
import pandas as pd
from polyfit import PolynomRegressor, Constraints
from num2words import num2words
from scipy.optimize import curve_fit
import itertools
import copy
import matplotlib.pyplot as plt
import warnings

def VTNA_orders(data,VTNA_selection,order_range=[-1.5,2.5],resolution=8,deg=5,fit_metric='RMSE',iterations=6,constraint='monotonic',
                score_interval=0.15,fixed_order_species=None,initial_mesh_denser=True):
        """
        Calculates the overlay score across different reaction order values for each normalised species in the \
        VTNA_selection dictionary for the kinetic data "data". First, a uniform order grid is defined by the order_range and \
        resolution arguments. Then, VTNA_omissions is applied to remove experiments or datapoints within experiments \
        as defined in VTNA_selection. Next, the average output species concentration profiles are normalised to \
        1. Then, numpy arrays containing the data required for variable time normalisation using the trapezoid \
        rule is defined with padding to ensure compatibility with numpy and inputed to explore_orders() to calculate \
        the overlay score for each order combination by calculating the normalised time axis with VTNA_tT() \
        and performing a global fit of each time normalised output concentration profile with a fitting method defined by \
        keyword arguments "deg" and "constraint". Explore_orders() then returns the overlay scores, each order combination \
        investigated and the best order combination overall. Next, another order grid with values around each minimum order value \
        is defined and explored by explore_orders() which calculates the overlay score for each combination. The new minimum \
        point is then used to define an even more fine order combination grid which is again given to explore_orders(). This \
        is repeated "iterations" number of time to improve the precision of the assigned order values. In the end, a \
        pandas DataFrame conatining the order values for each normalised species and the resulting overlay scores is \
        returned with the best order values and the intervals within which the overlay score deviates from the optimum \
        value by a factor of less than the "score_interval" value and the best order values found.
        Args:
            data (dict): Kinetic data as a dictionary of Pandas dataframes.
            VTNA_selection (dict): Dictionary containing information about how to carry out automatic VTNA.
            order_range (list): Minimum and maximum order value used to define the initial order value grid. 
            resolution (int): Number of datapoints included in the order value grids in each explore_orders() iteration.
            deg (int): Polynomial degree used in the global fitting prodecure to calculate the overlay score. 
            fit_metric (str): Goodness-of-fit metric chosen to define the overlay score value. Either 'R2', 'RMSE' or 'SE'. \
            otherwise variance is used.
            iterations (int): The number of times explore_orders() is called after the first order grid has been investigated.
            constraint (str): Defines the constraint of the global fit used by explore_orders() to calculate the overlay score. \
            Can be set to 'monotonic' for monotonic fitting (default), None for ordinary polynomial fitting or 'origin' for \
            linear fit through origin.
            score_interval: Defines the cutoff minimum_overlay_score + minimum_overlay_score*score_interval (max for R2) for \
            quoting the range of order values that give overlay scores close to that of the best order values to quantify how \
            well defined best order values are. 
            fixed_order_species (list or None): List containing the normalised reaction species for which \
            time axis should be normalised with a fixed value (as defined in the VTNA selection dictionary). \
            This can be done to lower the dimensionality of the overlay score versus order matrix. 
            initial_mesh_denser (bol): Determines whether the initial order value mesh should be denser than the \
            resolution argument by a factor of 1.5 (True) or the same as the resolution parameter (False).
        """
        # Check that the constraint argument is specified correctly.
        if constraint!=None and constraint!='origin' and constraint!='monotonic':
            raise ValueError("Constraint can only be set to 'origin' for linear fit through origin, 'monotonic' for \
                          monotonic fit or None for ordinary polynomial fit")
        # Check the intial time:
        start_time=time.time()
        print('Auto-VTNA calculation underway.')
        # Check if the number of fixed order species is greater than the number of normalised reaction species.
        if fixed_order_species!=None:
            no_fixed_order_species=len(fixed_order_species)
            if type(fixed_order_species)==str:
                no_fixed_order_species=1
            if no_fixed_order_species>=len(VTNA_selection['normalised_species'].items()):
                raise ValueError("The number of fixed order species can't be the same as or lower than the \
    number of normalised reaction species in the VTNA selection dictionary.")
        # Obtain deepcopy of the kinetic data and VTNA selection dictionary to avoid alterations. 
        data_copy=copy.deepcopy(data)
        selection_copy=copy.deepcopy(VTNA_selection)
        # Access one of the experimental dataframes.
        example_RM=list(data_copy.keys())[0]
        # Define the time axis key as t.
        t=data_copy[example_RM].keys()[0] 
        # Use the selected resolution and order range to define a list of order combinations. 
        # Resultion is increased by 50% for the first order exploration.
        if initial_mesh_denser:
            resolution_scaler=1.5
        else:
            resolution_scaler=1
        orders=np.linspace(order_range[0],order_range[1],round(resolution*resolution_scaler))
        # Edit the kinetic data according to the VTNA selection dictionary and add a scaled output values column.
        data_omissions=VTNA_omissions(data_copy,selection_copy)
        data_scaled=data_scaling_y(data_omissions,selection_copy)
        # Collect the time, normalised species and output species concentration data into designated lists.
        output_lists=[]
        NS_lists=[]
        time_lists=[]
        normalised_species=list(selection_copy['normalised_species'].keys())
        # Check that the reaction species listed in 'fixed_order_species' are present as normalised reaction 
        # species in the VTNA selection dictionary. Remove these reaction species from the normalised_species list. 
        if fixed_order_species!=None:
            # Make fixed_order_species into a list if only one reaction species string is defined. 
            if type(fixed_order_species)==str:
                fixed_order_species=[fixed_order_species]
            # Cancel calculation if fixed order reaction species have been selected that are not defined
            # as normalised reaction species in the VTNA selection dictionary. 
            if all(species in normalised_species for species in fixed_order_species)!=True:
                raise ValueError("All reaction species listed in 'fixed_order_species' must be present as a normalised \
                species in the selection dictionary")
            # Remove the fixed reaction species from the normalised_species list.
            normalised_species=[i for i in normalised_species if i not in fixed_order_species]
            # Generate a list of the fixed order values from the VTNA selection dictionary. 
            fixed_orders=[selection_copy['normalised_species'][i] for i in fixed_order_species]
            NS_lists_fixed_orders=[]
        # Find the number of normalised species and add a duplicate of orders to order_combinations for each.
        no_orders=len(normalised_species)
        order_combinations=[]
        for i in range(no_orders):
            order_combinations.append(orders)
        # Obtain the names of the experiment datasets in the data dictionary.
        experiments=list(data_copy.keys())
        # If selected experiments are defined in the selection dictionary, use these experiment names instead.
        if 'RM' in list(selection_copy.keys()):
            experiments=list(selection_copy['RM'].keys())
        # Extract the information required for time variable normalisation and fitting across order combinations. 
        counter=0
        for i in normalised_species:
            NS_nested_list=[]
            for exp in experiments:
                # Obtain the concentration profiles for each of the normalised reaction species i from each experiment exp.
                NS_nested_list.append(list(data_scaled[str(exp)][i].to_numpy()))
                # Collect the time and scaled output concentration values for each experiment once. 
                if counter==0:
                    time_lists.append(list(data_scaled[str(exp)][t].to_numpy()))
                    output_lists.append(list(data_scaled[str(exp)]['output_scaled'].to_numpy()))
            counter=1
            NS_lists.append(NS_nested_list)
        # Extract the integral factors from the kinetic data for the normalised reaction species with fixed order values. 
        if fixed_order_species!=None:
            counter=0
            for i in fixed_order_species:
                NS_nested_list_fixed_orders=[]
                for exp in experiments:
                    # Obtain the concentration profiles for each of the normalised reaction species i from each experiment exp.
                    NS_nested_list_fixed_orders.append(list(data_scaled[str(exp)][i].to_numpy()))
                NS_lists_fixed_orders.append(NS_nested_list_fixed_orders)
        # Find the number of time points for the kinetic data of each experiment.
        dimensions=[len(l) for l in time_lists]
        # Make the time and normalised species concentration lists from each experiment the same length by padding with 1.
        # Padding is done with 1 rather than 0 to avoid warning messages from trying to raise 0 to negative powers. 
        if all(l==dimensions[0] for l in dimensions)==False:
            max_dim=max(dimensions)
            # Add 0 to the end of each time list which is shorter than the longest list.
            time_lists = [sublist + [1] * (max_dim - len(sublist)) for sublist in time_lists]
            # For each nested list containing the concentration profiles for a given normalised species across all experiments, perform padding. 
            for nl in range(len(NS_lists)):
                NS_lists[nl]=[sublist + [1] * (max_dim - len(sublist)) for sublist in NS_lists[nl]]
            if fixed_order_species!=None:
                # For each nested list containing the concentration profiles for a given normalised species with fixed reaction orders 
                # across all experiments, perform padding. 
                for nl in range(len(NS_lists_fixed_orders)):
                    NS_lists_fixed_orders[nl]=[sublist + [1] * (max_dim - len(sublist)) for sublist in NS_lists_fixed_orders[nl]]
        # Now that the time and normalised species lists are padded, they are converted into 2d and 3d arrays respectively. 
        NS_lists=np.array(NS_lists)
        # Use the convolve function for each 2d array in the 3d array to obtain the average normalised species raised to their respective order values.
        dNS_lists=np.apply_along_axis(lambda x: np.convolve(x, [0.5,0.5], mode='valid'), axis=-1, arr=NS_lists)
        time_lists=np.array(time_lists)
        # Obtain the dt values for each time list.
        dt_lists=np.diff(np.array(time_lists), axis=1)
        # Multiply each dt with the integral fastors from the fixed reaction order reaction species. 
        if fixed_order_species!=None:
            NS_lists_fixed_orders=np.array(NS_lists_fixed_orders)
            # Turn zeros into very small values to avoid errors. 
            dNS_lists_fixed_orders=np.apply_along_axis(lambda x: np.convolve(x, [0.5,0.5], mode='valid'), axis=-1, arr=NS_lists_fixed_orders)
            order_array=np.array(fixed_orders)[:, np.newaxis,np.newaxis]
            # Raise the concentration values of each experiment dataset's normalised species. 
            dNS_lists_fixed_orders_updated=np.power(dNS_lists_fixed_orders, order_array)
            # Multiply together the 2d normalised species arrays and then multiply with the dt values.
            dt_lists=np.apply_along_axis(lambda x: np.prod(x, axis=0), axis=0, arr=dNS_lists_fixed_orders_updated)*dt_lists
        # Access the output concentration data of the first experiment to judge if increasing or decreasing monotonic fitting should be applied. 
        fit_direction='inc'
        if (output_lists[0][0]-output_lists[0][-1])>0:
            fit_direction='dec'
        # Perform first order exploration using "orders". 
        exploration1=explore_orders(dt_lists,output_lists,dNS_lists,dimensions,order_combinations,fit_metric,deg,constraint,fit_direction)
        best_orders=exploration1[-1]
        print(f'Best reaction order(s) from the 1st order exploration: {best_orders}')
        # Save the order combination tuples and overlay scores obtained in the first order exploration.
        FQs=exploration1[0]
        order_combinations=exploration1[1]
        # The zoom factor is defined as 140% of the difference between adjacent orders in the first reaction order grid applied if resolution>5. 
        # If the resolution is 4 or 5, the zoom factor is defined as the 110% and 120% of the difference between adjacent orders.  
        mesh_resolution=abs(orders[0]-orders[1])
        multipliers = {x: 1.1 if x == 4 else 1.2 if x == 5 else 1.4 for x in range(4, 10000)}
        zoom_factor=mesh_resolution*multipliers[resolution]
        # Perform a number of additional order explorations given by "interations" to increase the precision of the optimal orders. 
        for i in range(iterations):
            # Define resolution 2 to alternate between odd and even number of values in order value mesh.
            if i//2*2==i:
                resolution2=resolution+1
            else:
                resolution2=resolution
            orders_new=[]
            # Define a new set of order combinations arround the optimal order values from the first order exploration. 
            for k in range(no_orders):
                orders_new.append(np.linspace(best_orders[k]-zoom_factor,best_orders[k]+zoom_factor,resolution2))
            # Perform new order exploration of the new order value combinations. 
            exploration=explore_orders(dt_lists,output_lists,dNS_lists,dimensions,orders_new,fit_metric,deg,constraint,fit_direction)
            best_orders=exploration[-1]
            print(f'Best reaction order(s) from the {num2words(i+2, to="ordinal_num")} order exploration: {best_orders}')
            # Add datapoints from the latest order exploration to the FQ and order_combinations list. 
            FQs=FQs+exploration[0]
            order_combinations=order_combinations+exploration[1]
            # Define the new zoom factor as before to increase the fineness of the subsequent order value grid. 
            mesh_resolution=abs(exploration[1][-1][-1]-exploration[1][-2][-1])
            zoom_factor=mesh_resolution*multipliers[resolution+1 if resolution2==resolution else resolution]
        # Obtain the minimum point order combination (minium FQ for variance, RMSE and SE, max FQ for R2).
        minFQ=min(FQs)
        if fit_metric=='R2':
            minFQ=max(FQs)
        index_best_order=FQs.index(minFQ)
        best_orders_3=order_combinations[index_best_order]
        # Identify order combinations with overlay scores up to "score_interval" higher than that of the best order:
        if fit_metric=='R2':
            good_order_combos=[order_combinations[i] for i in range(len(order_combinations)) if FQs[i]>minFQ*(1-score_interval)]
        else: 
            good_order_combos=[order_combinations[i] for i in range(len(order_combinations)) if FQs[i]<minFQ*(1+score_interval)]
        # Identify the smallest and largest order values in the good order combinations list for each normalised species.
        largest_values=[max(column) for column in zip(*good_order_combos)]
        smallest_values=[min(column) for column in zip(*good_order_combos)]
        intervals=list(zip(smallest_values, largest_values))
        intervals=np.array(list(zip(smallest_values, largest_values)))
        interval_dict={'Normalised species':normalised_species,f'Lower order limit {score_interval*100}%':intervals[:,0],f'Upper order limit {score_interval*100}%':intervals[:,1]}
        # Create Dataframe for overlay scores against reaction order values.
        result_dict={}
        order_combinations_array=np.array(order_combinations)
        for i in range(len(normalised_species)):
            result_dict[f'order in {normalised_species[i]}']=order_combinations_array[:,i]
        result_dict[f'{fit_metric} overlay score']=FQs
        # Check the final time and print the time taken for the calculation
        end_time=time.time()
        total_time=end_time-start_time
        print(f'The calculation is complete. Time elapsed: {round(total_time,1)} seconds')
        # If the fit degree is set to 0, find the slope of the fit when optimal order values are applied. 
        # The x and y axis are not scaled.
        slope=None
        if deg==1:
            counter=0
            # Update the selection dictonary with the best reaction order values. 
            for species in normalised_species:
                if type(fixed_order_species)==list:
                    if species in fixed_order_species:
                        continue
                selection_copy["normalised_species"][species]=float(best_orders_3[counter])
                counter+=1
            output_species=selection_copy['output_species']
            tables=VTNA_new(data_copy,selection_copy)
            Data_points_y=[]
            Data_points_t=[]
            # Collect all the non-scaled datapoints from the normal VTNA calculation. 
            for i in tables.keys():
                x_values=[float(j) for j in tables[i]['tT'].to_numpy()]
                Data_points_t=Data_points_t+x_values
                ys=[float(j) for j in tables[i][output_species].to_numpy()]
                Data_points_y=Data_points_y+ys
            # Perform the linear fit to determine the slope of the fitted line.  
            if constraint=='origin':
                params = curve_fit(origin_line, Data_points_t, Data_points_y)
                slope=params[0][0]
            else:
                polyfit=np.polyfit(Data_points_t,Data_points_y,deg=deg)
                slope=polyfit[0] 
        return pd.DataFrame(result_dict),best_orders_3,pd.DataFrame(interval_dict),slope

def explore_orders(dt_lists,output_lists,dNS_lists,dimensions,orders,fit_metric,deg,constraint,fit_direction):
    """
    Calculates the normalised time axis of each order combination by calling VTNA_tT() with the \
    condensed kinetic data NS_lists and time_lists. Then performs the specified global fitting \
    procedure to calculate the overlay score as the goodness of fit measure "fit_metric". Returns \
    the overlay scores, order combinations and the order combination giving the best overlay score. 
    Args:
        time_lists (list of lists): Padded list containing time points from each experiment.
        output_lists (list of lists): List containing lists of output concentration profiles from each experiment. 
        NS_lists (list of list of lists): List containing the lists of padded concentration profile lists for \
        each normalised reaction species for each experiment.
        Dimensions (list of int): The number of rows in the DataFrame of each experiment in "data". Used to remove padded values. 
        orders (list of numpy arrays): List of numpy arrays containing the order values to be investigated for each \
        normalised reaction species.
        fit_metric (str): Goodness-of-fit metric chosen to define the overlay score value. Either 'R2', 'RMSE' or 'SE'. \
        otherwise variance is used.
        deg (int): Polynomial degree used in the global fitting prodecure to calculate the overlay score. 
        constraint (str): Defines the constraint of the global fit used by explore_orders() to calculate the overlay score. \
        Can be set to 'monotonic' for monotonic fitting (default), None for ordinary polynomial fitting or 'origin' for \
        linear fit through origin.
        fit_direction (str): The direction of the monotonic fitting procedure, either "inc" or "dec" as determined in \
        VTNA_orders()
    """
    # Nested function for calculating the normalised time axis
    if constraint=='monotonic':
        # Prepare "polyestimator" for performing monotonic fit with selected degree.
        polyestimator = PolynomRegressor(deg=deg, regularization = None, lam = 0)
        monotone_constraint = Constraints(monotonicity=fit_direction)
    # Define accumulator for goodness of fit measures.
    FQs=[]
    # Get every combination of orders as list of tuples.
    order_pairs=list(itertools.product(*orders))
    # Collapse the output list of lists to use in fitting.
    y=list(itertools.chain(*output_lists))
    for i in order_pairs:
        # Obtain the normalised time axis by calling VTNA_tT().
        tT=VTNA_tT(dt_lists,dNS_lists,i).tolist()
        # conncatenate the list of transformed time axis lists while removing padded values using the dimensions list 
        tT=list(itertools.chain(*[tT[i][:dimensions[i]] for i in range(len(dimensions))]))
        # Perform the polynomial fit of "deg" order using the specified fit method and evaluate fitted function at the relevant "tT" values. 
        # Then evaluate the goodness-of-fit by the selected fit metric and add it to the FQs list.
        if constraint=='origin':
            tT_2=[i/max(tT) for i in tT]
            if deg==1:
                params = curve_fit(origin_line, tT_2, y)
                a=params[0]
                y_pred=a*tT_2
                a
            else:
                print('for fit through zero, use deg=1')
                return 
            FQ=score_fit(y,y_pred,fit_metric)
            FQs.append(FQ)
        if constraint=='monotonic': 
            tT_2=np.array([i/max(tT) for i in tT]).reshape((-1,1))    
            polyestimator.fit(tT_2, y, loss = 'l2', constraints={0: monotone_constraint})       
            pred_y = polyestimator.predict(tT_2)
            # Call the score_fit function to evalue the goodness-of-fit using the fit measurement "fit_metric".
            FQ=score_fit(y,pred_y,fit_metric)
            FQs.append(FQ)
        if constraint==None:
            tT_2=[i/max(tT) for i in tT]
            polyfit=np.polyfit(tT_2,y,deg=deg)
            pred_y=np.polyval(polyfit,tT_2)
            FQ=score_fit(y,pred_y,fit_metric)
            FQs.append(FQ)
    # Obtain the best fit metric value (min for SE, RMSE and variance, max for R2)
    minFQ=min(FQs)
    if fit_metric=='R2':
        minFQ=max(FQs)
    # Identify the reaction orderds corresponding to the order minimum point
    index_best_order=FQs.index(minFQ)
    best_orders=order_pairs[index_best_order]
    return FQs, order_pairs, best_orders

def VTNA_tT(dt_lists,dNS_lists,order_combination):
        """
        Calculates the numerical integral of the product of the concentrations of normalised species over time \
        raised to their respective order values in order_combination. The numerical integrals are calcualted using the trapezoid rule. 
        Args: 
            time_lists, NS_lists: see explore_orders() 
            order_combination (tuple): the reaction order values used to calculate the normalised time axis. 
        """
        # Make an order combination matrix to match the normalised species 3d array.
        order_array=np.array(order_combination)[:, np.newaxis,np.newaxis]
        # Raise the concentration values of each experiment dataset's normalised species. 
        dNS_lists_updated=np.power(dNS_lists, order_array)
        # Multiply together the 2d normalised species arrays and then multiply with the dt values.
        tT_lists=np.apply_along_axis(lambda x: np.prod(x, axis=0), axis=0, arr=dNS_lists_updated)*dt_lists
        # Obtain the trapezoid rule integrals for each time point using the numpy cumsum function.
        tT_summed=np.cumsum(tT_lists,axis=1)
        # Add 0 to the beginning of each transformed time axis array.
        tT_summed = np.apply_along_axis(lambda x: np.insert(x, 0, 0), axis=1, arr=tT_summed)
        return tT_summed

def data_scaling_y(data,selection):
    """
    Creates a new column in the DataFrames in the kinetic data dictionary "data" consisting of output \
    concentration profiles shifted to start at the same concentration value (only concequential if \
    a reactant rather than product is defined as the output species) and scaled to ensure an average \
    output concentration of 1 across every row of every experiment DataFrame. 
    Args:
            data (dict): Kinetic data as a dictionary of Pandas dataframes.
            VTNA_selection (dict): Dictionary containing information about how to carry out automatic VTNA.
    """
    output_species=selection['output_species']
    VTNA_data=copy.deepcopy(data)    
    species_initial={}
    # Loop through every chosen kinetic profile to collect the first output species datapoint from each experiment.
    if 'RM' in selection.keys():
        experiments=list(selection['RM'].keys())
        for i in selection['RM'].keys():
            species_initial[i]=VTNA_data[i][output_species].to_numpy()[0]
    # If no kinetic profiles have been chosen, find the largest concentration of the output species
    # from all the kinetic profiles in the dataset.
    else: 
        experiments=list(VTNA_data.keys())
        for i in VTNA_data.keys():
            species_initial[i]=VTNA_data[i][output_species].to_numpy()[0]
    #Use largest output concentration value to scale all output concentrations in the VTNA dataset
    max_FOV=max([species_initial[i] for i in list(species_initial.keys())])   
    # Save scaled output concentration values as new column for each reaction profile
    output_values=[]
    for i in experiments:
        shift_value=max_FOV-species_initial[i]
        VTNA_data[str(i)]['output_scaled']=VTNA_data[i][output_species].to_numpy()+shift_value
        output_values=output_values+list(VTNA_data[str(i)]['output_scaled'].to_numpy())
    # Find the average output value across all experiments and use this to scale the 'output_scaled' columns.
    scaling_value=np.mean(output_values)
    for i in experiments:     
        VTNA_data[str(i)]['output_scaled']=np.array([float(i) for i in VTNA_data[i]['output_scaled']])/scaling_value
    return VTNA_data

def on_click(event,VTNA_selection,data,y_unit,size_scaler,fit_metric,deg,constraint,fixed_order_species=None,axis_species=None,show_legend=True):
    """
    Creates a VTNA overlay plot for the kinetic dataset "data" according to the VTNA selection dictionary \
    "VTNA_selection" with order values corresponding to the coordinates of the user's click on a graph or contour \
    plot generated with plot_orders_vs_overlay(). 
    Args:
        event: Contains the x and y positions of the user's click.
        VTNA_selection (dict): Dictionary containing information on the automatic VTNA calculation which generated \
        the overlay score versus order graph or contour plot. 
        data (dict): Kinetic data as a dictionary of Pandas dataframes.
        deg (int): Polynomial degree used in the global fitting prodecure to calculate the overlay score. 
        fit_metric (str): Goodness-of-fit metric chosen to define the overlay score value. Either 'R2', 'RMSE' or 'SE'. \
        otherwise variance is used.
        constraint (str): Defines the constraint of the global fit used by explore_orders() to calculate the overlay score. \
        Can be set to 'monotonic' for monotonic fitting (default), None for ordinary polynomial fitting or 'origin' for \
        linear fit through origin. 
        fixed_order_species (list or None): List containing the normalised reaction species for which \
        time axis should be normalised with a fixed value (as defined in the VTNA selection dictionary). \
        This can be done to lower the dimensionality of the overlay score versus order matrix. 
        axis_species (list): The normalised species of each axis of the overlay score versus order plot. \
        For two normalised species, the x-axis species will appear before the y-axis species. 
        show_legend (bol): Can be set to False to not show the legend of the overlay plot generated. 
    """
    VTNA_selection_copy=copy.deepcopy(VTNA_selection)
    output_species=VTNA_selection_copy['output_species']
    normalised_species_list=list(VTNA_selection_copy['normalised_species'].keys())
    # Create a list of the experiments to include in the fitting procedure. 
    experiments=list(data.keys())
    # If selected experiments RM are defined in the selection dictionary, use these experiment names instead.
    if 'RM' in list(VTNA_selection_copy.keys()):
        experiments=list(VTNA_selection_copy['RM'].keys())
    # Define x and y coordinates as the reaction orders and update the VTNA selection accordingly
    orders=[event.xdata,event.ydata]
    # Define the number of decimals to be included in the x-axis label.
    # Generate data for the second plot based on the click coordinates
    # Define x and y coordinates as the reaction orders and update the VTNA selection accordingly
    siphers=[]
    x_label="$\u03A3"
    counter=0
    if type(fixed_order_species)==str:
        fixed_order_species=[fixed_order_species]
    for i in axis_species:
        VTNA_selection_copy['normalised_species'][i]=orders[counter]
        counter=counter+1
    for j in normalised_species_list:
        q=5
        if str(VTNA_selection_copy['normalised_species'][j])[0]=='-':
            q=q+1
        siphers.append(q)
        x_label=x_label+f"[{j}]^{{{str(VTNA_selection_copy['normalised_species'][j])[:q]}}}"
    x_label=x_label+"\u0394t$."
    # Generate scaled VTNA table for orders.
    tables=VTNA_new(data,VTNA_selection_copy)
    # Scale the output concentration profiles.
    tables_scaled=data_scaling_y(tables,VTNA_selection_copy)
    # Create string to add to plot title if some normalised species have fixed order values
    title_string=f'VTNA overlay plot for {output_species}'
    if fixed_order_species!=None:
        if type(fixed_order_species)==str:
            title_string=title_string+f"\nFixed order for {fixed_order_species}: {VTNA_selection_copy['normalised_species'][fixed_order_species]}"
        else:
            for i in fixed_order_species:
                title_string=title_string+f"\nFixed order for {i}: {VTNA_selection_copy['normalised_species'][i]}. "
    fig2, ax2 = plt.subplots(figsize=(5*size_scaler,4*size_scaler))
    # Check if the output concentration profiles start at the same values for each experiment.
    if 'RM' in list(VTNA_selection.keys()):
        initial_output_values=[data[RM][VTNA_selection['output_species']].to_numpy()[0] for RM in list(VTNA_selection['RM'].keys())]
    else: 
        initial_output_values=[data[RM][VTNA_selection['output_species']].to_numpy()[0] for RM in list(data.keys())]
    if all(elem==initial_output_values[0] for elem in initial_output_values):
        shifted=False
    else:
        shifted=True
    if event.button==3 or event.button==2:
        # Check difference between first and last output values to judge whether fit polynomial should 
        # be monotonically increasing or decreasing.
        delta_output=0
        for RM in experiments:
            if (tables_scaled[RM]['output_scaled'][0]-tables_scaled[RM]['output_scaled'].to_numpy()[-1])>0:
                delta_output=delta_output-1
            else: 
                delta_output=delta_output+1
        fit_direction='dec'
        if delta_output>0:
            fit_direction='inc'
        polyestimator = PolynomRegressor(deg=deg, regularization = None, lam = 0)
        monotone_constraint = Constraints(monotonicity=fit_direction)
        # Plot the time-normalised reaction profiles for each experiment dataset in tables
        Data_points_y=[]
        Data_points_t=[]
        times=[]
        for i in experiments:
            times.append(max([float(j) for j in tables_scaled[i]['tT'].to_numpy()]))
        max_time=max(times)
        for i in experiments:
            # Define the x and y values. 
            x_values=[float(j)/max_time for j in tables_scaled[i]['tT'].to_numpy()]
            Data_points_t=Data_points_t+x_values
            ys=[float(j) for j in tables_scaled[i]['output_scaled'].to_numpy()]
            Data_points_y=Data_points_y+ys
            # Plot the datapoints of the concentration profile 
            ax2.plot(x_values,ys,marker='o',label=i)  
        # Set teh y and x axis labels of the plot. 
        ax2.set_ylabel(f'[{output_species}] / {y_unit}')
        ax2.set_xlabel(x_label)
        # Show the legend if show_legend is set to True. 
        if show_legend:
            legend_1=ax2.legend()
            ax2.add_artist(legend_1)
        y_label=f'Scaled [{output_species}]'
        if shifted:
            y_label='Shifted '+y_label
        ax2.set_ylabel(y_label)
        if constraint=='monotonic':
            # Perform the polynomial fit of "deg" order and evaluate fitted function at the relevant "tT" values       
            Data_points_t_2=np.array(Data_points_t).reshape((-1,1))
            polyestimator.fit(Data_points_t_2, Data_points_y, loss = 'l2', constraints={0: monotone_constraint})       
            pred_mon = polyestimator.predict(Data_points_t_2)
            ts_2=np.linspace(0,1,100).reshape((-1,1))
            pred_mon_fit = polyestimator.predict(ts_2)
            fit_plot,=ax2.plot(ts_2,pred_mon_fit,marker='',label=f'Monotonic {deg}. \norder polynomial fit',linestyle='dotted',linewidth=2,color='black')
            # Call the score_fit function to evalue the goodness-of-fit using the fit measuremetn "fit_metric"
            FQ=score_fit(Data_points_y,pred_mon,fit_metric)
        elif constraint=='origin':
            params = curve_fit(origin_line, Data_points_t, Data_points_y)
            a=params[0]
            y_pred=a*Data_points_t
            FQ=score_fit(Data_points_y,y_pred,fit_metric)
            ts_2=np.linspace(0,max(Data_points_t),100)
            fit_plot,=ax2.plot(ts_2,ts_2*a,marker='',label='linear fit\nvia origin',linestyle='dotted',linewidth=2,color='black')
        else:
            if constraint not in [None,'None']:
                warnings.warn("Ordinary polynomial fitting applied due to unknown constraint input", UserWarning)
            polyfit=np.polyfit(Data_points_t,Data_points_y,deg=deg)
            pred_y=np.polyval(polyfit,Data_points_t)
            tT_2=np.linspace(0,1,100)
            pred_y_fit=np.polyval(polyfit,tT_2)
            FQ=score_fit(Data_points_y,pred_y,fit_metric)
            fit_plot,=ax2.plot(tT_2,pred_y_fit,marker='',label=f'{deg}. order\npolynomial fit',linestyle='dotted',linewidth=2,color='black')
        FQ_round=round(FQ,6)
        plt.title(title_string+f'\nGOF ({fit_metric}): {FQ_round}')
        if show_legend:
            ax2.legend(handles=[fit_plot],loc='lower center')
        plt.tight_layout()
        plt.show()
    elif event.button==1:
        # This corresponds to a left-click.
        # Plot the time-normalised reaction profiles for each experiment dataset in tables
        for i in experiments:
            x_values=[float(j) for j in tables_scaled[i]['tT'].to_numpy()]
            ys=[float(j) for j in tables_scaled[i][output_species].to_numpy()]
            ax2.set_ylabel(f'[{output_species}] / {y_unit}')
            ax2.plot(x_values,ys,marker='o',label=i)
        if show_legend:
            ax2.legend()
        # Define the number of decimals to be included in the x-axis label.
        q,w=5,5
        if str(orders[0])[0]=='-':
            q=1+q
        if str(orders[-1])[0]=='-':
            w=1+w
        # Define the axis labels and the title of the overlay plot. 
        ax2.set_xlabel(x_label)
        y_label=f'Concentration in {y_unit}'
        if shifted:
            y_label='Shifted '+y_label
        ax2.set_ylabel(y_label)
        plt.title(title_string)
        plt.tight_layout()
        plt.show()

def score_fit(y, y_fit, fit_metric):
    # NB: this function was written with the help of OpenAI's ChatGTP. 
    # Find the variance of the residuals of the fit
    FQ = np.var(np.array(y_fit) - np.array(y))
    # Update the goodness-of-fit value according to the chosen fit metric.
    if fit_metric == 'R2':
        # Compute R-squared manually
        SS_res = np.sum((np.array(y) - np.array(y_fit))**2)
        SS_tot = np.sum((np.array(y) - np.mean(np.array(y)))**2)
        FQ = 1 - (SS_res / SS_tot)
    elif fit_metric == 'SE':
        # Compute Standard Error of the Estimate
        FQ = FQ**0.5 / len(y)**0.5
    elif fit_metric == 'RMSE':
        # Compute Root Mean Squared Error
        FQ = np.sqrt(np.mean((np.array(y_fit) - np.array(y))**2))
    return FQ

def VTNA_new(data,VTNA_selection):
    """
    First removes experimental data as specified in "VTNA_selection" from "data". Then \
    calculates the normalised time axis using the trapezoid rule for the concentration \
    profiles of the reaction species listed in "normalised_species" with their respective \
    order values.
    Args: 
        data (dict): Kinetic data as a dictionary of Pandas dataframes.
        VTNA_selection (dict): Dictionary containing information about how to carry out automatic VTNA.
    """
    VTNA_table={}
    #Access one of the experimental dataframes
    example_RM=list(data.keys())[0]
    #Define the time axis key as t
    t=data[example_RM].keys()[0] 
    first_output_values={}
    first_output_values2=[]
    output_species=VTNA_selection['output_species']    
    if 'RM' in list(VTNA_selection.keys()):
        #loop through every selected RM dataframe
        for RM in VTNA_selection['RM'].keys():
            #initialise list TN for normalised time axis
            TN=[float(data[RM][t][0])]
            table=data[RM]
            #Apply omissions of selected datapoints from RM dataframe
            if VTNA_selection["RM"][RM]['omissions']!=None:
                if VTNA_selection["RM"][RM]['omissions'][0]=='range':
                    table=table.drop([c for c in range(len(table)) if c/len(table)*100>VTNA_selection["RM"][RM]['omissions'][1] and c/len(table)*100<VTNA_selection["RM"][RM]['omissions'][2]])
                else: 
                    for j in VTNA_selection["RM"][RM]['omissions']:
                        table=table.drop([int(j)])
            #Update the numbers of each row in case a row has been omitted so that next step works
            table=table.set_index([pd.Index([i for i in range(len(table[t]))])])
            #run through the list of data points for an experiment
            for i in range(len(table[t])-1):
                C=1
                #Define dt as time difference between time point i+1 and the previous one i
                dt=float(table[t][i+1])-float(table[t][i])
                #Apply numerical integration of each normalised species selected in VTNA_selection
                for NS in VTNA_selection["normalised_species"].keys():
                    order=float(VTNA_selection["normalised_species"][NS])
                    C=C*((float(table[NS][i+1])+float(table[NS][i]))/2)**order
                #Add the C trapezoid to the overall normalised time axis
                TN.append(TN[i]+dt*C)
            #Define the running time integral TN as the transformed time axis tT
            table['tT']=TN
            VTNA_table[RM]=table
            # Append the first output concentration value to the first output list and dictionary           
            first_output=float(table[output_species].to_numpy()[0])
            first_output_values[RM]=first_output
            first_output_values2.append(first_output)   
    #Code applied if no RM selections are specified:
    if 'RM' not in list(VTNA_selection.keys()):
        #Update the numbers of each row in case a row has been omitted so that next step works
        for d in data.keys(): 
            TN=[0]
            table=data[d]
            table=table.set_index([pd.Index([i for i in range(len(table[t]))])])
            for i in range(len(table[t])-1):
                #running through the list of data points for an experiment
                C=1
                dt=float(table[t][i+1])-float(table[t][i])
                #Apply numerical integration of the normalised species in VTNA_selection dict
                for NS in VTNA_selection["normalised_species"].keys():
                    order=float(VTNA_selection["normalised_species"][NS])
                    C=C*((float(table[NS][i+1])+float(table[NS][i]))/2)**order
                #Add the dC trapezoid to the overall integrated concentration integral
                TN.append(TN[i]+dt*C)
            table['tT']=TN
            VTNA_table[d]=table
            # Append the first output concentration value to the first output list and dictionary
            first_output=float(table[output_species].to_numpy()[0])
            first_output_values[d]=first_output
            first_output_values2.append(first_output)
    # Find the reaction profile with the largest initial concentration of the output species
    max_FOV=max(first_output_values2)
    # Shift the output reaction profiles to ensure identical starting points.
    for RM in VTNA_table.keys():
        shift_value=max_FOV-first_output_values[RM]
        VTNA_table[RM][output_species]=VTNA_table[RM][output_species].to_numpy()+shift_value
    return VTNA_table

def origin_line(x,a):
    # Function used for performing linear fit through origin
    return x*a

def simple_line(x,a,b):
    # Function used for performing linear fit through origin
    return x*a+b

def VTNA_omissions(data,VTNA_selection,range_mode="Percentage"):
    """
    Uses the 'RM' section of VTNA_selection dictionary to remove experiment DataFrames\
    or selected concentration values within  experiment DataFrames. 
    Args:
            data (dict): Kinetic data as a dictionary of Pandas dataframes.
            VTNA_selection (dict): Dictionary containing information about how to carry out automatic VTNA.
    """
    VTNA_table=copy.deepcopy(data)
    if 'RM' in list(VTNA_selection.keys()):
        VTNA_table={}
        #Access one of the experimental dataframes
        example_RM = next(iter(data.keys()))
        #Define the time axis key as t
        t=data[example_RM].keys()[0] 
        #loop through every selected RM dataframe
        for RM in VTNA_selection['RM'].keys():
            #initialise list TN for normalised time axis
            TN=[float(data[RM][t][0])]
            table=data[RM]
            max_t=table[t].to_numpy()[-1]
            #Apply omissions of selected datapoints from RM dataframe
            if VTNA_selection["RM"][RM]['omissions']!=None:
                if VTNA_selection["RM"][RM]['omissions'][0]=='range':
                    if range_mode=="Percentage":
                        table=table.drop([c for c in range(len(table)) if table[t].to_numpy()[c]
                        /max_t*100>VTNA_selection["RM"][RM]['omissions'][1] and table[t].to_numpy()[c]/max_t*100<=VTNA_selection["RM"][RM]['omissions'][2]])
                    else: 
                        table=table.drop([c for c in range(len(table)) if table[t].to_numpy()[c]
                        >VTNA_selection["RM"][RM]['omissions'][1] and table[t].to_numpy()[c]<=VTNA_selection["RM"][RM]['omissions'][2]])
                else: 
                    for j in VTNA_selection["RM"][RM]['omissions']:
                        table=table.drop([int(j)])
            table=table.set_index([pd.Index([i for i in range(len(table[t]))])])
            VTNA_table[RM]=table
    return VTNA_table


from auto_vtna.VTNA_functions import VTNA_orders, on_click
import numpy as np
import pandas as pd
import itertools
from functools import partial
import matplotlib.pyplot as plt
from openpyxl.drawing.image import Image
from io import BytesIO
import matplotlib.ticker as tkr
import warnings

class Automatic_VTNA:
    def __init__(self,data,VTNA_selection,order_range=[-1.5,2.5],resolution=10,deg=5, \
                 fit_metric='SE',iterations=4,constraint='monotonic',score_interval=0.15,\
                fixed_order_species=None,initial_mesh_denser=True):
        # Initialise the class:
        self.data=data.copy()
        self.VTNA_selection=VTNA_selection.copy()
        self.order_range=order_range
        self.resolution=resolution
        self.deg=deg
        self.fit_metric=fit_metric
        self.iterations=iterations
        self.constraint=constraint
        self.score_interval=score_interval
        self.fixed_order_species=fixed_order_species
        self.initial_mesh_denser=initial_mesh_denser
        # Perform automatic VTNA calculation:
        self.results, self.best_orders, self.interval, self.best_slope = VTNA_orders(self.data,self.VTNA_selection, \
        self.order_range,self.resolution,self.deg,self.fit_metric,self.iterations,self.constraint, \
        self.score_interval,self.fixed_order_species,self.initial_mesh_denser)

    def identify_error_intervals(self,score_interval=0.15):
        """
        Generates a Pandas dataframe containing the lowest and highest order value which is found in the \
        list of order value combinations that give an overlay score within "score_interval" of the optimal overlay \
        score. The score_interval can be set to any value above 0. 
        """
        # Get the column names except the rightmost one
        col_names = self.results.columns[:-1]
        # Create an empty list to store lists of column values
        combined_values = [[] for _ in range(len(self.results))]
        # Iterate through each row of the DataFrame
        for index, row in self.results.iterrows():
            # Iterate through each column except the rightmost one
            for col_name in col_names:
                # Append the value of each column to the corresponding list
                combined_values[index].append(row[col_name])
        # Convert the list of lists into a numpy 2D array
        order_combinations = np.array(combined_values)
        # Extract the rightmost column as a numpy array
        FQs = self.results.iloc[:, -1].values
        # Obtain the minimum point order combination (minium FQ for variance, RMSE and SE, max FQ for R2).
        minFQ=min(FQs)
        if self.fit_metric=='R2':
            minFQ=max(FQs)
        # Identify order combinations with overlay scores up to "score_interval" higher than that of the best order:
        if self.fit_metric=='R2':
            good_order_combos=[order_combinations[i] for i in range(len(order_combinations)) if FQs[i]>minFQ*(1-score_interval)]
        else: 
            good_order_combos=[order_combinations[i] for i in range(len(order_combinations)) if FQs[i]<minFQ*(1+score_interval)]
        # Identify the smallest and largest order values in the good order combinations list for each normalised species.
        largest_values=[max(column) for column in zip(*good_order_combos)]
        smallest_values=[min(column) for column in zip(*good_order_combos)]
        intervals=list(zip(smallest_values, largest_values))
        intervals=np.array(list(zip(smallest_values, largest_values)))
        # Define a list of the normalised reaction species that haven't been fixed. 
        if type(self.fixed_order_species)!=list:
            fixed_order_list=[self.fixed_order_species]
        else:
            fixed_order_list=self.fixed_order_species
        # Define the list of normalised species and generate the overlay score interval dictionary. 
        normalised_species=[i for i in list(self.VTNA_selection['normalised_species'].keys()) if i not in fixed_order_list]
        interval_dict={'Normalised species':normalised_species,f'Lower order limit {score_interval*100}%':intervals[:,0],f'Upper order limit {score_interval*100}%':intervals[:,1]}
        return pd.DataFrame(interval_dict)

    def plot_orders_vs_overlay(self,points=False,size_scaler=1,size_scaler2=1,title=None,zoom=False,grid=False,color_scaler=1,
    fixed_cbar_resolution=None,interval=False,y_unit='M',title_fontsize=12,overlay_score_range_max=None,
    overlay_score_range_min=None,show_legend_main=True,show_legend_popup=True,decimals_in_colorbar=3,annotate_optimum=True,
    specified_score_interval=None):
        """
        Creates a graph or contour plot visualising the overlay score versus order value matrix \
        for one and two normalised reaction species respectively. If the overlay score versus \
        order value matrix involves 3 or more normalised species, no plot is created. The graph \
        or contour plots can be clicked upon to generate the VTNA overlay plot corresponding to \
        the order values of the click. 
        Args:
            points (bol): Set to True to visualise the order value data points for which overlay \
            scores have been determined. 
            size_scaler (int or float): Factor by which the size of the order value versus overlay \
            score plot is adjusted. 
            size_scaler2 (int or float): Factor by which the size of the VTNA overlay plot generated \
            by clicking on the original plot is adjusted. 
            title (str): The title of the order value versus overlay score plot.
            zoom (int or float): Width of order values around the optimial order point that will be \
            included in the order value versus overlay score plot.
            grid (bol): Set to True to include a grid in the order value versus overlay score plot.
            color_scaler (int or float, ∈<0,1]): Number between 1 and 0 which determines the factor by \
            which the contour plot color bar range is truncated. NB: This will give white sections that \
            are off the contour plot color map.
            resolution_contour (float): Difference in overlay score between separate color shades in \
            the contour plot. 
            interval (bol): Set to True to show each order combination datapoint within the interval_score \
            times the best overlay score value away from the optimal order value point. 
            y_unit (str): The concentration unit of the output concentration profiles. Gets included in the \
            yaxis title. 
            title_fontsize (int): Fontsize of the title of the order value versus overlay score plot.
            overlay_score_range_max (None or float): The maximum overlay score value to be shown in the plot \
            either as a contour plot color bar (2 NS) or the y axis of an ordinary graph (1 NS).
            overlay_score_range_min (None or float): The minimum overlay score value to be shown in the plot \
            either as a contour plot color bar (2 NS) or the y axis of an ordinary graph (1 NS).
            show_legend_main (bol): Set to True to show the legend of the overlay score versus order plot. 
            show_legend_popup (bol): Set to True to show the legend of the VTNA overlay plots generated by \
            mouseclicks on the overlay score versus order plot. 
            decimals_in_colorbar (int): The number of decimals to be shown in the colorbar of an overlay score \
            versus order contour plot. 
            annotate_optimum (bol): Set to True to show the order values of the optimum point of an overlay score \
            versus order contour plot. 

        """
        #if color_scaler>1 or color_scaler<=0:
        #    raise ValueError('color_scaler should be set to ∈<0,1]')
        # Find the species corresponding to the x or x and y axis:
        axis_species=[i.replace('order in ','') for i in list(self.results.columns)[:-1]]
        output_species=self.VTNA_selection['output_species']
        analysis_columns=self.results.columns
        no_normalised_species=len(analysis_columns)-1
        # Create string to include information about reaction species with fixed order value(s) during the 
        # automatic VTNA calculation.
        fixed_order_string=''
        if self.fixed_order_species!=None:
            if type(self.fixed_order_species)==str:
                fixed_order_string=fixed_order_string+f"\nFixed order for {self.fixed_order_species}: {self.VTNA_selection['normalised_species'][self.fixed_order_species]}"
            else:
                for i in self.fixed_order_species:
                    fixed_order_string=fixed_order_string+f"\nFixed order for {i}:{self.VTNA_selection['normalised_species'][i]}."
        if interval==True:
            # Identify the score interval from the calculation or from the specified_score_interval argument.
            score_interval=self.score_interval
            if specified_score_interval!=None:
                if float(specified_score_interval)>0:
                    score_interval=float(specified_score_interval)
                else:
                    warnings.warn("The specified score interval must be a number above 0. The calculation value will be used instead.")
        if no_normalised_species>2:
            return ValueError('The overlay score : order value matrix can only be visualized for 1 or 2 varying normalised reaction species.')
        if no_normalised_species==1:
            x_best=self.best_orders[0]
            # Define x values as orders and y values as fit metric values.
            x_s=np.array(self.results[analysis_columns[0]])
            y_s=np.array(self.results[analysis_columns[1]])
            # Sort the order values from smallest to largest to get one continuous plt line. 
            xy=list(itertools.zip_longest(x_s,y_s))
            xy.sort()
            # Redefine sorted x and y values
            y=[y[1] for y in xy]
            x=[x[0] for x in xy]
            # Find the finest orders interval used to find minimum point
            resolution_orders=str(format(x_s[-1]-x_s[-2],".12f"))
            # Find the index for first non 0 digit in resolution
            for i in range(len(resolution_orders)):
                if resolution_orders[i]!='.' and resolution_orders[i]!='0':
                    break
            # Increase i by 1 if first character in order is '-'
            if str(x_best)[0]=='-':
                i=i+1
            # Create figure with size scaled by size_scaler
            fig, ax = plt.subplots(figsize=(6*size_scaler,4*size_scaler))
            # Plot the fit metric values obtained at each order.
            ax.plot(x,y,marker='o',linestyle='-',markersize=5)
            # Plot the minimum point and insert label information for legend.
            minimum_info=f'Best order: {f"%.{i-1}f" % round(x_best,i-1)}\nUncertainty: {f"%.{i}f" % round(float(resolution_orders[:(i+2)]),i)}\nMinimum {self.fit_metric}: {f"%.{i+2}f" % round(min(y),i+2)}'
            optimum_value=min(y)
            if self.fit_metric=='R2':
                minimum_info=f'Best order: {f"%.{i-1}f" % round(x_best,i-1)}\nUncertainty: {f"%.{i}f" % round(float(resolution_orders[:(i+2)]),i)}\nMaximum {self.fit_metric}: {f"%.{i+2}f" % round(max(y),i+2)}'
                optimum_value=max(y)
            minimum_info=minimum_info+fixed_order_string
            plt.plot(x_best,optimum_value,color='orange',marker='o',label=minimum_info)
            # Define the x and y axis labels. 
            ax.set_xlabel(f'{analysis_columns[0]}',fontsize=12)
            ax.set_ylabel(f'Overlay score: {self.fit_metric}',fontsize=12)
            # Change the x- and y-axis limits in case zoom has been applied. 
            if zoom!=False:
                x1=x_best-zoom/2
                x2=x_best+zoom/2
                plt.xlim(x1,x2)
                y2=[i for i in y if x[y.index(i)]<x2 and x[y.index(i)]>x1]
                y1=min(y2)
                y2=max(y2)
                leeway=(y2-y1)*0.1
                ax.set_ylim(y1-leeway,y2+leeway)
            # Plot a horizontal line showing the datapoints within the interval_score*best overlay score range of the minimum point.
            if interval==True:
                interval_limit=min(y)+min(y)*score_interval
                if self.fit_metric=='R2':
                    interval_limit=max(y)-max(y)*score_interval
                ax.plot([self.order_range[0],self.order_range[1]],[interval_limit,interval_limit],linestyle='--',label=f"{float(score_interval)*100}% from optimal point.")
            # Set the figure title. 
            if title==None:
                fig.suptitle(f'VTNA overlay score versus orders for the formation of {output_species}',fontsize=title_fontsize,y=0.95)
            else:
                fig.suptitle(title,fontsize=title_fontsize,y=0.95)
            # Apply legend and grid if selected.
            if show_legend_main:
                ax.legend()
            if grid==True:
                plt.grid()
            # Define the ylimits if overlay score range max or min have been defined. 
            if overlay_score_range_max!=None:
                ax.set_ylim(top=overlay_score_range_max)
            if overlay_score_range_min!=None:
                ax.set_ylim(bottom=overlay_score_range_min)
        # Code for contour plot to illustrate the overlay score across order values if there are 2 normalised species. 
        if no_normalised_species==2:
            # Obtain the order value arrays x_s and y_s and the overlay scores z_s.
            z_s=np.array(self.results[analysis_columns[2]])
            x_s=np.array(self.results[analysis_columns[0]])
            y_s=np.array(self.results[analysis_columns[1]])
            # Define the x and y values of the minimum point (or maximum for R2).
            x_best=self.best_orders[0]
            y_best=self.best_orders[1]
            # Create figure with size scaled by size_scaler.
            fig, ax = plt.subplots(figsize=(6*size_scaler,5*size_scaler))
            # Identify the max and min fit metric values for setting up color grid for contour plot.
            min_z=min(z_s)
            max_z=max(z_s)
            # Find the finest orders interval used to find minimum point.
            resolution_orders=str(format(y_s[-1]-y_s[-2],".12f"))
            # Find the index for first non 0 digit in resolution_orders.
            for i in range(len(resolution_orders)):
                if resolution_orders[i]!='.' and resolution_orders[i]!='0':
                    break
            # Define q and w to determine number of digits to show for each minimum point order.
            q,w=i,i
            # Increase q and w by 1 if first character in respective order is '-'
            if str(x_best)[0]=='-':
                q=1+i
            if str(y_best)[-1]=='-':
                w=1+i
            if points==True:
                # Plot the data points used to generate the countour plot.
                ax.plot(x_s,y_s,'or',markersize=0.5)
            # Plot the datapoints with an overlay score off the minimum point (or max for R2) score.
            
            if interval==True:
                if self.fit_metric!='R2':
                    x_interval=[x_s[i] for i in range(len(x_s)) if z_s[i]<(score_interval+1)*min(z_s)]
                    y_interval=[y_s[i] for i in range(len(y_s)) if z_s[i]<(score_interval+1)*min(z_s)]
                else:
                    x_interval=[x_s[i] for i in range(len(x_s)) if z_s[i]>(-score_interval+1)*min(z_s)]
                    y_interval=[y_s[i] for i in range(len(y_s)) if z_s[i]>(-score_interval+1)*min(z_s)]
                # Plot the identified interval datapoints in green.
                ax.plot(x_interval,y_interval,markersize=0.7,linestyle='',marker='x',color='b',  
                        label=f'Orders within {score_interval*100}% of optimal O.S.',zorder=2)
            # Print axis labels with species_order from data to assign normalised species.
            ax.set_xlabel(f'{analysis_columns[0]}')
            ax.set_ylabel(f'{analysis_columns[1]}')
            # Plot the data point corresponding to the best VTNA alignment and use label to provide information.
            if self.fit_metric=='R2':
                ax.plot(x_best,y_best,marker='x',color='k',linestyle='',label = f'Best orders: ({round(x_best, q-1):.{int(q-1)}f},{round(y_best, w-1):.{int(w-1)}f})\nUncertainty: {resolution_orders[:(i+2)]}\nMaximum {self.fit_metric}: {round(max(z_s), i+2):.{int(i+2)}f}{fixed_order_string}')
            else:
                ax.plot(x_best,y_best,marker='x',color='k',linestyle='',label=f'Best orders: ({round(x_best, q-1):.{int(q-1)}f},{round(y_best, w-1):.{int(w-1)}f})\nUncertainty: {resolution_orders[:(i+2)]}\nMinimum {self.fit_metric}: {round(min(z_s), i+2):.{int(i+2)}f}{fixed_order_string}')
            # If zoom argument applied, apply xlim and ylim around the minimum point and find the lowest z_value in that interval.
            min_z=min(z_s)
            max_z=max(z_s)
            z_range=max_z-min_z
            max_z=max(z_s)-z_range*(1-color_scaler)
            # If the overlay score range max or min arguments have been defined, prepare the relevant variables.
            cbar_fixing=False
            if overlay_score_range_max!=None:
                cbar_fixing=True
                max_z=overlay_score_range_max
            if overlay_score_range_min!=None:
                cbar_fixing=True
                min_z=overlay_score_range_min
            if zoom!=False:
                if cbar_fixing==True:
                    warnings.warn("overlay_score_range_max and overlay_score_range_min is ignored when zoom isn't set to False.")
                # Define x and y limits to zoom in closer to the minimum point
                min_z=min(z_s)
                ax.set_xlim(x_best-zoom/2,x_best+zoom/2)
                ax.set_ylim(y_best-zoom/2,y_best+zoom/2)
                # Obtain the z_s values in the new domain.
                z_s2=[i for i in z_s if x_s[list(z_s).index(i)]<(x_best+zoom/2) and x_s[list(z_s).index(i)]>(x_best-zoom/2) \
                       and y_s[list(z_s).index(i)]<(y_best+zoom/2) and y_s[list(z_s).index(i)]>(y_best-zoom/2)] 
                # Define the maximum fit metric value in this interval and the difference between this and the minimum value
                max_z=max(z_s2)
                z_range=max_z-min_z
                # Use color_scaler to define to adjust the maximum value of the colorbar
                max_z=max_z-z_range*(1-color_scaler)
            # Define the order value range of the plot
            delta_order=max(x_s)-min(x_s)
            # Define appropriate colorbar values if R2 is used as fit metric
            if self.fit_metric=='R2':
                max_z=max(z_s)
                max_z=max_z+(1-max_z)*0.2
                min_z=min_z+z_range*(1-color_scaler)
            if fixed_cbar_resolution!=None:
                resolution_cbar=float(fixed_cbar_resolution)
            else:
                resolution_cbar=abs((max_z-min_z)/30)
            clev = np.arange(min_z,max_z,resolution_cbar)
            # Create contour plot.
            CS = ax.tricontourf(x_s, y_s, z_s, clev,cmap=plt.cm.viridis_r)
            if self.fit_metric=='R2':
                CS = ax.tricontourf(x_s, y_s, z_s, clev,cmap=plt.cm.viridis)
            cbar=fig.colorbar(CS,orientation='vertical',label=f'Overlay score: {self.fit_metric}',ax=ax,format=tkr.FormatStrFormatter(f'%.{decimals_in_colorbar}g'))
            legend_loc='best'
            if zoom==False:
                best_value_x=x_best+0.045
                best_value_y=y_best+0.045
                if best_value_y-min(y_s)<best_value_y-max(y_s):
                    legend_loc='upper '
                else: 
                    legend_loc='lower '
                if best_value_x-min(x_s)<best_value_x-max(x_s):
                    legend_loc=legend_loc+'right'
                else:
                    legend_loc=legend_loc+'left'
            if show_legend_main:
                ax.legend(loc=legend_loc)
            # Define the title of the figure.
            if title==None:
                fig.suptitle(f'VTNA overlay score versus orders for the formation of {output_species}',fontsize=title_fontsize,y=0.95)
            else:
                fig.suptitle(title,fontsize=title_fontsize,y=0.95)
            # Annotate the minimum point to show its order values. 
            # Adjust position according to the "zoom" value and delta_order if zoom isn't set to False
            if annotate_optimum:
                if zoom!=False:
                    ax.annotate(f'{f"%.{q}f" % round(x_best,q)}, {f"%.{w}f" % round(y_best,w)}',(x_best+0.045*zoom/delta_order, y_best+0.045*zoom/delta_order))
                if zoom==False:
                    ax.annotate(f'{f"%.{q}f" % round(x_best,q)}, {f"%.{w}f" % round(y_best,w)}',(best_value_x,best_value_y))
        # Create a new click function with inputs.
        on_click_with_inputs = partial(on_click,data=self.data,VTNA_selection=self.VTNA_selection,y_unit=y_unit,\
        size_scaler=size_scaler2,fit_metric=self.fit_metric,deg=self.deg,constraint=self.constraint,fixed_order_species=self.fixed_order_species,
        axis_species=axis_species,show_legend=show_legend_popup)
        # Bind the on_click function to the mouse click event.
        cid = fig.canvas.mpl_connect('button_press_event', on_click_with_inputs)
        # Show the plot.
        plt.tight_layout()
        if grid==True:
            plt.grid()
        image_stream = BytesIO()
        plt.savefig(image_stream, format='png', bbox_inches='tight', pad_inches=0)
        image_stream.seek(0)
        img = Image(image_stream)
        plt.show()
        return img

